unit uHouse;

interface

uses
  uPattern                  // for TVisited
  , classes                 // for TList
  , graphics                // for TColor
  ;

type

  TMaterial = (tmWood, tmPlastic);

  THouse = class(TVisited)
  private
    FWindows: TList;
    FDoors: TList;
    FTop: integer;
    FLeft: integer;
    FHeight: integer;
    FWidth: integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AcceptVisitFrom(Visitor: TVisitor); override;
    property Top: integer read FTop write FTop;
    property Left: integer read FLeft write FLeft;
    property Height: integer read FHeight;
    property Width: integer read FWidth;
    property Windows: TList read FWindows;
    property Doors: TList read FDoors;
  end;

  THoleStopper = class(TVisited)
  private
    FHouse: THouse;
    FMaterial: TMaterial;
    FColour: TColor;
    FTop: integer;
    FLeft: integer;
    FHeight: integer;
    FWidth: integer;
  protected
    function GetStopperType: string; virtual;
  public
    constructor Create; virtual;
    procedure AcceptVisitFrom(Visitor: TVisitor); override;
    property House: THouse read FHouse write FHouse;
    property Material: TMaterial read FMaterial write FMaterial;
    property Colour: TColor read FColour write FColour;
    property Top: integer read FTop write FTop;
    property Left: integer read FLeft write FLeft;
    property Height: integer read FHeight write FHeight;
    property Width: integer read FWidth write FWidth;
    property StopperType: string read GetStopperType;
  end;

  TWindow = class(THoleStopper)
  protected
    function GetStopperType: string; override;
  public
    constructor Create; override;
  end;

  TDoor = class(THoleStopper)
  protected
    function GetStopperType: string; override;
  public
    constructor Create; override;
  end;

implementation

{ THouse }

procedure THouse.AcceptVisitFrom(Visitor: TVisitor);
var
  nStopper: integer;
begin
  // Let the visitor do something to the whole house
  Visitor.AllowVisitorToDoSomethingTo(Self);
  // Let the visitor visit the windows
  for nStopper := 0 to Pred(Windows.Count) do
    THoleStopper(Windows.Items[nStopper]).AcceptVisitFrom(Visitor);
  // Let the visitor visit the doors
  for nStopper := 0 to Pred(Doors.Count) do
    THoleStopper(Doors.Items[nStopper]).AcceptVisitFrom(Visitor);
end;

constructor THouse.Create;
var
  wnd: TWindow;
  door: TDoor;
  nhs: integer;         // short for nHoleStopper
begin
  inherited;
  // Set the house's dimensions
  FHeight := 400;
  FWidth := 500;
  // Create the list of windows
  if not Assigned(FWindows) then
    FWindows := TList.Create;
  // Add 4 windows
  FWindows.Clear;
  for nhs := 0 to 3 do
  begin
    wnd := TWindow.Create;
    wnd.Height := 80;
    wnd.Width := 100;
    wnd.Left := 100 + (200 * (nhs mod 2));
    wnd.Top := 80 + (160 * (nhs div 2));
    wnd.House := Self;
    FWindows.Add(wnd);
  end;
  // Create the list of doors
  if not Assigned(FDoors) then
    FDoors := TList.Create;
  // Add 2 doors
  FDoors.Clear;
  for nhs := 0 to 1 do
  begin
    door := TDoor.Create;
    door.Height := 160;
    door.Width := 80;
    door.Left := 210 + (205 * (nhs mod 2));
    door.Top := Self.Height - door.Height - 5;
    door.House := Self;
    FDoors.Add(door);
  end;
end;

destructor THouse.Destroy;
var
  nhs: integer;
begin
  // Destroy the windows
  for nhs := 0 to Pred(FWindows.Count) do
    TWindow(FWindows.Items[nhs]).Free;
  FWindows.Free;
  // Destroy the doors
  for nhs := 0 to Pred(FDoors.Count) do
    TDoor(FDoors.Items[nhs]).Free;
  FDoors.Free;
  // Finish the house demolition
  inherited;
end;

{ THoleStopper }

procedure THoleStopper.AcceptVisitFrom(Visitor: TVisitor);
begin
  Visitor.AllowVisitorToDoSomethingTo(Self);
end;

constructor THoleStopper.Create;
begin
  inherited;
  FMaterial := tmWood;
end;

function THoleStopper.GetStopperType: string;
begin
  Result := 'Something to stop the rain getting in';
end;

{ TWindow }

constructor TWindow.Create;
begin
  inherited;
  Self.Colour := clWhite;
end;

function TWindow.GetStopperType: string;
begin
  Result := 'A Window';
end;

{ TDoor }

constructor TDoor.Create;
begin
  inherited;
  Self.Colour := clBlack;
end;

function TDoor.GetStopperType: string;
begin
  Result := 'A Door';
end;

end.
